/* WSPacket -- Wrapper around data kept by Burp for each websockets message.
 *
 * Copyright (c) 2016 Eddy Vanlerberghe.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of Eddy Vanlerberghe shall not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EDDY VANLERBERGHE ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL EDDY VANLERBERGHE BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
package evilwan;

import java.util.*;
import java.lang.reflect.*;

/**
 * This class does not expose all the data that is recorded by Burp for websocket
 * messages, just those that seemed to be the most relevant,
 * <p>
 * Note that this class requires that at least one websocket message has been
 * recorded by Burp before initializing the various static <code>Method</code> fields.
 */
public class WSPacket {
    // {{ Class members }}

    /**
     * Flag indicating whether or not to dump extra debug info for reversing
     * the Burp internal WS packet class.
     */
    private static boolean  _show_privates = false;
    //
    // Note: class names have been found via clever reverse engineering, in
    // combination with "educated guess work"
    //
    // Methods have to be located at run-time (need to have references to
    // Method objects and those methods have to be searched based on signature
    // as well as return type)
    //
    /**
     * Burp websocket message encapsulation class
     */
    private static Class	_wsclazz = burp.qeh.class;
    /**
     * Generic message data encapsulation class
     */
    private static Class	_dataclazz = burp.rjb.class;
    /**
     * Burp byte[] encapsulation class
     */
    private static Class	_barclazz = burp.ald.class;
    /**
     * Burp URL encapsulation class
     */
    private static Class	_earlclazz = burp.ejd.class;
    /**
     * Accessor method on _barclazz instances
     */
    private static Method	_burp2bar = null;
    /**
     * Accessor methods on _wsclazz instances. 
     */
    private static Method	_get_comment = null;
    private static Method	_get_time = null;
    private static Method	_get_c2s = null;
    private static Method	_get_url = null;
    private static Method	_get_orgmsg = null;
    private static Method	_get_cgdmsg = null;
    
    // {{ Instance members }}
    
    /**
     * Encapsulated Burp websocket message object
     */
    private Object _burp_packet = null;

    // {{ Class methods }}

    /**
     * Helper method for printing messages.
     * <p>
     * The printed message is written to <code>System.out</code> and starts with the current
     * class name, followed by two dashes, followed by the specified text.
     * @param s contains the <code>String</code> value to print.
     */
    private static void say(String s) {
	System.out.println("WSPacket -- " + s);
    }

    /**
     * Look for only method in byte array encapsulation class that returns a byte[].
     *
     * @param c byte array encapsulation class
     * @return method that will return the actual data of the encapsulated array
     * as a byte array
     */
    private static Method getBarExtraction(Class c) {
	Method m = null;
	Method[] mets = c.getDeclaredMethods();
	for(int i = 0; i < mets.length; ++i) {
	    Class[] paramtypes = mets[i].getParameterTypes();
	    if(paramtypes.length == 0) {
		if(mets[i].getReturnType().equals(byte[].class)) {
		    m = mets[i];
		    break;
		}
	    }
	}
	return m;
    }
    //
    // Code based on debugging as well as reverse engineering of parts of Burp code.
    //
    /**
     * Find which methods to use for extracting properties of the objects that
     * encapsulate one single intercepted websocket message.
     * <p>
     * Methods are identified both by name (signature) and their return type. The
     * obfuscation tool used by Portswigger apparently creates multiple instances
     * of methods with the same signature (name + parameter type list) but with
     * different return types: in Java class files, methodref items contain not only
     * the signature part of a method (name and types of arguments) but also the
     * return type, so that is why the obfuscator gets away with it.
     *
     * @param c class providing encapsulation of all websocket message related data.
     * For version 1.7.04 of Burp, this is class burp.qeh
     */
    private static void getAccessorMethods(Class c) {
	Method[] mets = c.getDeclaredMethods();
	for(int i = 0; i < mets.length; ++i) {
	    Class[] paramtypes = mets[i].getParameterTypes();
	    //
	    // Only interested in "accessor" methods: methods taking no arguments
	    // and returning a specific property from the object
	    //
	    if(paramtypes.length == 0) {
		if(mets[i].getReturnType().equals(long.class)) {
		    //
		    // long is typically used for timestamps and there is only one
		    // long property, so this must be the equivalent of getTime()
		    //
		    _get_time = mets[i];
		} else if(mets[i].getReturnType().equals(boolean.class)) {
		    //
		    // There is only one boolean property, so this must be the one
		    //
		    _get_c2s = mets[i];
		} else if(mets[i].getReturnType().equals(_dataclazz) &&
			  mets[i].getName().equals("b")) {
		    _get_orgmsg = mets[i];
		} else if(mets[i].getReturnType().equals(String.class) &&
			  mets[i].getName().equals("a")) {
		    _get_comment = mets[i];
		} else if(mets[i].getReturnType().equals(_dataclazz) &&
			  mets[i].getName().equals("a")) {	// v1.6.05
		    _get_cgdmsg = mets[i];
		} else if(mets[i].getReturnType().equals(_earlclazz)) {
		    //
		    // Only one accessor method returning this type...
		    //
		    _get_url = mets[i];
		}
	    }
	}
    }

    //
    // Toggle flag that will dump additional info useful for reversing
    // the Burp internal representation of a WS packet
    //
    public static void setDebug(boolean flag) {
	_show_privates = flag;
    }
    public static boolean getDebug() {
	return _show_privates;
    }
    //
    // Get remaining info from Burp environment for working with websocket messages.
    //
    // This includes extracting accessor methods for various websocket message
    // properties, as well as accessor method for extracting the raw message bytes
    // (it appears that the TableModel ArrayList does not store byte[] directly but
    // instead uses another (or more?) encapsulation layer(s)
    //
    // For this remaining analysis we start by using info that can be obtained from
    // the first element in the table (such as class and accessor methods)
    //
    // After executing this method we have all the info we need to dump the
    // websocket messages.
    //
    /**
     * Initialize static method references for various accessor methods.
     * <p>
     * Note that for Burp version 1.7.04 the single argument to this method
     * is of type burp.pn
     *
     * @param datalist <code>TableModel</code> containing the websocket message
     * data ... somewhere
     */
    public static void init(Object datalist) {
	try {
	    Object[] al = ((Collection) datalist).toArray();
	    if(al.length <= 0) {
		throw new Exception("WebSocket table data model is empty.");
	    }
	    Object o = al[0];	// get first element from that list for further investigation
	    //
	    // If commandline option "-XXX" (without the quotes) is given to Burp then
	    // print out the name of the class that represents an internal WS packet and then dump
	    // the static string info in that class (those are de-obfuscated by the
	    // classloader, so they contain human readable info by now) Those de-obfuscated texts
	    // can then be used to link obfuscated property names and methods to items like original message
	    // contents, comments etc.
	    //
	    // Feed that output to the accompanying Perl script that will then print out the new method names
	    // that are required below.
	    //
	    if(getDebug()) {
		say("init() -- _wsclazz is: " + _wsclazz.getName());
		Foefel.dumpStatics(_wsclazz);
	    }
	    getAccessorMethods(_wsclazz);
	    _burp2bar = getBarExtraction(_barclazz);
	} catch(Throwable t) {
	    System.out.println("## caught: " + t);
	    t.printStackTrace();
	    System.exit(1);
	}
    }

    // {{ Constructors }}
    
    /**
     * Default public constructor.
     */
    public WSPacket(Object bp) {
	_burp_packet = bp;
    }
    
    // {{ Instance methods }}

    //
    // Accessor methods to Burp websocket message objects
    //
    /**
     * Return flag indicating if the message was sent from client (browser) to server or not.
     *
     * @return <code>true</code> if the message was sent from client to server
     */
    public boolean getClientToServer() {
	try {
	    return ((Boolean)_get_c2s.invoke(_burp_packet)).booleanValue();
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return false;
    }
    /**
     * Get time when the websocket message was intercepted
     *
     * @return time when the message was intercepted by Burp
     */
    public long getTimestamp() {
	try {
	    return ((Long)_get_time.invoke(_burp_packet)).longValue();
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return 0L;
    }
    /**
     * Get the comment text that was assigned to the message.
     *
     * @return comment text that is assigned to the message or
     * <code>null</code> if no comment was given
     */
    public String getComment() {
	try {
	    return (String)_get_comment.invoke(_burp_packet);
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return null;
    }
    /**
     * Get URL used for establishing the websocket.
     *
     * @return URL as a <code>String</code> used for creating the websocket
     */
    public String getURL() {
	try {
	    Object o = _get_url.invoke(_burp_packet);;
	    return _get_url.invoke(_burp_packet).toString();
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return null;
    }
    /**
     * Get the original data sent as a websocket message.
     *
     * @return original data payload of the websocket message
     */
    public byte[] getOriginalMessage() {
	try {
	    Object o = _get_orgmsg.invoke(_burp_packet);;
	    if(o == null) {
		return null;
	    } else {
		return (byte[])_burp2bar.invoke(o);
	    }
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	    say("getOriginalMessage() -- caught:" + iae);
	    iae.printStackTrace();
	}
	return null;
    }
    /**
     * Get modified payload of the websocket message.
     *
     * @return modified payload of the websocket message or <code>null</code>
     *  if the message was not intercepted and modified
     */
    public byte[] getEditedMessage() {
	try {
	    Object o = _get_cgdmsg.invoke(_burp_packet);;
	    if(o == null) {
		return null;
	    } else {
		return (byte[])_burp2bar.invoke(o);
	    }
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return null;
    }
}
